from functools import wraps

from flask import current_app

from ..utils import get_wrap_str, filter_list

__all__ = ['get_current_model_rest_manager_callback',
           'rest_login_required', 'rest_permission_required',
           'get_rest_log_msg',
           'log_operation',
           'gen_route_method']


def get_current_model_rest_manager_callback(callback_name):
    if current_app:
        model_rest_manager = getattr(current_app, 'model_rest_manager', None)
        if model_rest_manager:
            return getattr(model_rest_manager, callback_name, None)


def rest_login_required():
    """
    If you decorate a view with this, it will ensure that the current user is logged in and authenticated before calling the actual view.

    Example:
        @sys_mgmt_bp.route('/auth/account/', methods=['GET', 'POST'])
        @rest_login_required()
        def sys_auth_account_query():
            pass

    :return:
    """

    def decorate(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            check_result = True
            login_check_callback = get_current_model_rest_manager_callback('login_check_callback')
            if login_check_callback:
                check_result = login_check_callback()
            if check_result is not True:
                return check_result
            return func(*args, **kwargs)

        return wrapper

    return decorate


def rest_permission_required(module, action=None):  # @2023-10-10, op_permission--> action
    """
    If you decorate a view with this, it will ensure that the current user.
    has the module permission and action permission before calling the actual view.

    Example:
        @sys_mgmt_bp.route('/role/', methods=['GET'])
        @rest_permission_required('role')
        def sys_role_query():
            pass

        @sys_mgmt_bp.route('/role/', methods=['POST'])
        @rest_permission_required('role', 'add')
        def sys_role_add():
            pass

    :param module: module permission to be checked
    :param action: action permission to be checked
    :return:
    """

    def decorate(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            check_result = True
            permission_check_callback = get_current_model_rest_manager_callback('permission_check_callback')
            if permission_check_callback:
                check_result = permission_check_callback(module, action)
            if check_result is not True:
                return check_result
            return func(*args, **kwargs)

        return wrapper

    return decorate


def get_rest_log_msg(api_info, req_data, success, res_data):
    """
    Return the log message text.

    :param api_info:
    :param req_data:
    :param success:
    :param res_data:
    :return:
    """
    return get_wrap_str(
        '--' + api_info,
        '--Request data:', req_data,
        '--Response result:' + str(success),
        '--Response data:', res_data)


def log_operation(*args, **kwargs):
    """
    Log operation/action.

    Example:
        log_operation('users', 'login', success, request_json.get('username'), res_data)

    :param args:
    :param kwargs:
    :return:
    """
    logging_callback = get_current_model_rest_manager_callback('logging_callback')
    if logging_callback:
        logging_callback(*args, **kwargs)


def gen_route_method(method, url_prefix, view_func_name=None):
    """
    Generate endpoint unique function name.
    :param method:
    :param url_prefix:
    :param view_func_name:
    :return:
    """

    def decorator(f):
        def wrap(*args, **kwargs):
            return f(*args, **kwargs)

        if view_func_name is None:
            methods = url_prefix.split('/')
            methods.append(method)
            methods.insert(0, 'model_route')
            wrap.__name__ = '_'.join(filter_list(methods, lambda item: item != '')).replace('-', '_')
        else:
            wrap.__name__ = view_func_name
        return wrap

    return decorator



from contextlib import contextmanager

import sqlalchemy
from flask import g
from sqlalchemy import text, or_, and_, inspect

from . import DBSession
from ._base import BaseModelMixin
from ..utils import get_g_cache, set_g_cache, remove_g_cache, parse_version

__all__ = ['create_instance', 'create_relationships',
           'query_all_models', 'query_multiple_model',
           'append_debug_queries', 'get_debug_queries', 'append_query_filter',
           'get_db_session', 'db_session', 'close_db_session',
           'model_to_dict', 'refresh_instance',
           'is_model_mixin_instance'
           ]


def create_instance(model_cls, data, create_relationship=True):
    """
    Create an instance of the specified model class with the data.

    Example:
        create_instance(User, {"name": "taozh", "email": "taozh@focus-ui.com"})

    :param model_cls: The specified model class.
    :param data: The data used to create instance.
    :param create_relationship: If true, the relationships will be created.
    :return:
    """
    if not isinstance(data, dict):
        return None

    col_value_dict = model_cls.filter_attrs_by_columns(data)
    instance = model_cls(**col_value_dict)

    if create_relationship is True:
        relationships = create_relationships(model_cls, data)
        for key in relationships:
            setattr(instance, key, relationships[key])
    return instance


def create_relationships(model_cls, data):
    """
    Create the relationship dict of the specified model class with the data.

    :param model_cls:
    :param data:
    :return:
    """
    relationships = model_cls.get_relationships()
    relationship_map = {}
    for key in relationships.keys():
        relationship_cls = relationships[key].mapper.class_
        relationship_kwargs = data.get(key)
        if isinstance(relationship_kwargs, list):  # 1:n
            relationship = []
            for item in relationship_kwargs:
                r_ins = create_instance(relationship_cls, item)
                if r_ins is not None:
                    relationship.append(r_ins)
        else:
            relationship = create_instance(relationship_cls, relationship_kwargs)  # 1:1

        if relationship is not None:
            relationship_map[key] = relationship

    return relationship_map


def query_all_models(*models):
    """
    Query all the data of the specified model(ModelMixin) list.

    .. versionadded:: 1.5

    Example:
        result = query_all_models(User, Role)

    :param models:
    :return:
    """
    result = []
    for model in models:
        r = model.query_all()

        if type(r) is tuple:  # ModelMixin
            if r[0] is not True:  # error
                return r
            else:
                result.append(r[1])
        else:
            result.append(r)
    return result


def query_multiple_model(*cls_list):
    """
    Query all the data of the multiple specified model(ModelMixin) class.

    Example:
        result = query_multiple_model(User, Role)

    :param cls_list:
    :return:If failed, returns the failed tuple, otherwise, returns the the data list.
    """

    return query_all_models(*cls_list)


def append_debug_queries(query):
    if _has_g_context():  # @2022-07-26 add,
        debug_queries = getattr(g, 'z_debug_queries', None)
        if debug_queries is None:
            g.z_debug_queries = debug_queries = []
        debug_queries.append(query)


def get_debug_queries():
    if _has_g_context():  # @2022-07-26 add,
        return getattr(g, 'z_debug_queries', [])
    return []


def append_query_filter(query, filters, joined):  # @2023-06-21 add
    """
    Append filters to the query
    :param query:
    :param filters:
    :param joined:
    :return:
    """
    if len(filters) == 0:
        return query
    joined = joined.lower()
    if joined == 'or':
        joined_text = ' OR '
        joined_func = or_
    elif joined == 'and':
        joined_text = ' AND '
        joined_func = and_
    else:
        return query

    text_items = []
    binary_expression_items = []
    for item in filters:
        if type(item) is str:
            text_items.append(item)
        else:
            binary_expression_items.append(item)

    if len(text_items) > 0:
        query = query.filter(text('(' + (joined_text.join(text_items)) + ')'))

    if len(binary_expression_items) > 0:
        query = query.filter(joined_func(*binary_expression_items))

    return query


def get_db_session():
    """
    Get the db session from g(flask)/ Create a db session(without request).
    If not exist, create a session and return.

    Example:
        session = get_db_session()
        session.query(User).all()
        session.close()             # close session

    :return:
    """
    if _has_g_context():  # @2022-07-26 add, make sure work without flask request
        session = get_g_cache('_flaskz_db_session')
        if session is None:
            session = DBSession()
            set_g_cache('_flaskz_db_session', session)
    else:
        session = DBSession()
        setattr(session, '_temporary', True)  # @2022-08-22 add '_temporary' attr
    return session


def close_db_session():
    """
    Close the cached session.

    :return:
    """
    if _has_g_context():  # @2022-07-26 add
        session = get_g_cache('_flaskz_db_session')
        if session is not None:
            remove_g_cache('_flaskz_db_session')
            session.close()


def _close_temporary_session(session):
    """
    @2023-05-06: add, close temporary(non-cached) session
    :param session:
    :return:
    """
    if getattr(session, '_temporary', None) is True:
        session.close()


@contextmanager
def db_session(do_commit=True):
    """
    Database session context manager.

    Example:
        with db_session() as session:
            instance = create_instance(cls, json_data)
            session.add(instance)

    :param do_commit: If false, session will not commit,generally used for query operations
    :return:
    """
    session = get_db_session()
    try:
        yield session
        if do_commit is not False:
            session.commit()
    except Exception as e:
        if do_commit is not False:
            session.rollback()
        raise e
    _close_temporary_session(session)


@contextmanager
def _db_session(do_commit, do_rollback):
    """
    # @2023-08-21: add, for internal use
    """
    session = get_db_session()
    try:
        yield session
        if do_commit is True:
            session.commit()
    except Exception as e:
        if do_rollback is True:
            session.rollback()
        raise e
    _close_temporary_session(session)


def model_to_dict(ins, option=None):
    """
    Convert model data to dict.

    Example:
        result = Role.update(request_json)
        res_data = model_to_dict(result[1], {
            'cascade': 1
            # 'filter': lambda ins: ins.type == 'local'
        })

    :param ins:
    :param option:
    :return:
    """

    if isinstance(ins, list):
        data_list = []
        ins_filter = None
        if type(option) is dict:
            ins_filter = option.get('filter')
        if not callable(ins_filter):
            ins_filter = None
        for item in ins:
            if ins_filter and ins_filter(item) is not True:  # 2023-09-19: add
                continue
            if is_model_mixin_instance(item):  # @2022-11-28: change, ModelMixin --> BaseModelMixin
                data_list.append(item.to_dict(option))
            else:
                data_list.append(item)
        return data_list
    elif is_model_mixin_instance(ins):
        return ins.to_dict(option)
    else:
        return ins


def refresh_instance(ins):
    """
    Expire and refresh attributes on the given instance/list

    .. versionadded:: 1.6.4

    Example:
        user = User.query_by({'name': 'flaskz'}, True)
        # ...
        refresh_instance(user)

        user_list = User.query_by({'name': 'flaskz'})
        # ...
        refresh_instance(user_list)

    :param ins:
    :return:
    """
    if ins is None:
        return
    if not isinstance(ins, list):
        ins_list = [ins]
    else:
        ins_list = ins
    for item in ins_list:
        _refresh_instance(item)


def _refresh_instance(ins):  # @2023-10-17 add
    if not is_model_mixin_instance(ins):
        return
    ins_inspect = inspect(ins)
    ins_session = ins_inspect.session
    if ins_session and ins_session.is_active is True:
        ins_session.refresh(ins)


def is_model_mixin_instance(obj):
    """
    Check if the object is an instance of the BaseModelMixin.

    :param obj:
    :return:
    """
    return isinstance(obj, BaseModelMixin)


def _has_g_context():
    # if has_request_context():  # If there is request context, g must exist
    #     return True
    if not g:  # @2022-11-28: change, (not g) != (g is None)
        return False
    return True
    # return has_request_context() or g is not None


_sa_version = parse_version(sqlalchemy.__version__)
if _sa_version[0] > 2 or _sa_version[0] > 1 and _sa_version[1] >= 4:
    def _session_get(session, cls, ident):
        """Return an instance based on the given primary key identifier, or None if not found."""
        return session.get(cls, ident)
else:
    def _session_get(session, cls, ident):
        """Return an instance based on the given primary key identifier, or None if not found."""
        return session.query(cls).get(ident)



from collections.abc import Mapping

__all__ = [
    'filter_list', 'find_list', 'merge_list', 'each_list', 'get_list',
    'get_dict', 'del_dict_keys', 'get_deep', 'set_deep', 'merge_dict', 'get_ins_mapping', 'get_dict_mapping', 'get_dict_value_by_type',
    'is_list', 'is_dict', 'slice_str',
    'get_wrap_str', 'is_str', 'str_replace',
    'bulk_append_child',
    'parse_version'
]


# -------------------------------------------list-------------------------------------------
def filter_list(item_list, func, with_index=False):
    """
    Filter the list and return a new list that matches the conditions
    :param with_index:
    :param item_list:
    :param func:
    :return:
    """
    if with_index is True:
        return _filter_list_with_index(item_list, func)

    return list(filter(func, item_list))


def _filter_list_with_index(item_list, func):
    result = []
    for index, item in enumerate(item_list):
        if func(item, index) is True:
            result.append(item)
    return result


def find_list(item_list, func, with_index=False):
    """
    Find an object from the list that matches the conditions
    :param with_index:
    :param item_list:
    :param func:
    :return:
    """
    if with_index is True:
        return _find_list_with_index(item_list, func)

    for item in item_list:
        if func(item) is True:
            return item
    return None


def _find_list_with_index(item_list, func):
    for index, item in enumerate(item_list):
        if func(item, index) is True:
            return item
    return None


def merge_list(lst, *to_merged_list):
    """
    Merge multiple lists into the first list
    :param lst:
    :param to_merged_list:
    :return:
    """
    for item in to_merged_list:
        lst.extend(item)
    return lst


def each_list(item_list, func, with_index=True):
    """
    Iterate each object in the list and execute the callback function
    :param with_index:
    :param item_list:
    :param func:
    :return:
    """
    if with_index is True:
        return _each_list_with_index(item_list, func)

    for item in item_list:
        if func(item) is False:
            return


def _each_list_with_index(item_list, func):
    for index, item in enumerate(item_list):
        if func(item, index) is False:
            return


def get_list(obj, default=None):
    """
    Get the list instance
    -- If obj is a list, just return
    -- If obj is not a list, return default or []
    :param obj:
    :param default:
    :return:
    """
    return obj if is_list(obj) else (default or [])


# -------------------------------------------dict-------------------------------------------
def get_dict(obj, default=None):
    """
     Get the dict instance
    -- If obj is a dict, just return
    -- If obj is not a dict, return default or {}

    :param obj:
    :param default:
    :return:
    """
    return obj if is_dict(obj) else (default or {})


def del_dict_keys(obj, keys):
    """
    Delete the specified keys from the dict object
    :param obj:
    :param keys:
    :return:
    """
    for key in keys:
        obj.pop(key, None)


def get_deep(obj, key, key_split='.', default=None, raising=False):
    """
    Get the deep value of the specified key from the dict object
    :param obj:
    :param key:
    :param key_split:
    :param default:
    :param raising:
    :return:
    """
    value = obj
    try:
        for key in key.split(key_split):
            if isinstance(value, dict):
                value = value[key]
                continue
            else:
                if raising:
                    raise KeyError
                return default
    except KeyError:
        if raising:
            raise
        return default
    else:
        return value


def set_deep(obj, key, value, key_split='.'):
    """
    Set the deep value of the specified key from the dict object
    :param obj:
    :param key:
    :param value:
    :param key_split:
    :return:
    """
    dd = obj
    keys = key.split(key_split)
    latest = keys.pop()
    for k in keys:
        dd = dd.setdefault(k, {})
    dd[latest] = value


def get_ins_mapping(ins_list, attr, deep=False):
    """
    Create a dict map of the specified object list with the specified attribute
    :param ins_list:
    :param attr:
    :param deep:
    :return:
    """
    map_dict = {}
    if deep is True:
        keys = attr.split('.')
        for item in ins_list:
            k_value = item
            for k in keys:
                k_value = getattr(k_value, k)
            map_dict[k_value] = item
    else:
        for item in ins_list:
            map_dict[getattr(item, attr, None)] = item

    return map_dict


def get_dict_mapping(dict_list, key='id', key_join="+"):
    """
    Create a dict map of the specified dict list withe the specified key
    :param dict_list:
    :param key:
    :param key_join:
    :return:
    """
    map_dict = {}
    for item in dict_list:
        if isinstance(key, list):
            kv = []
            for k_item in key:
                kv.append(str(get_deep(item, k_item)))
            map_dict[key_join.join(kv)] = item
        else:
            map_dict[get_deep(item, key)] = item

    return map_dict


def merge_dict(dct, *merged_dict_list):
    """
    Recursive dict merge. Inspired by :meth:``dict.update()``, instead of
    updating only top-level keys, dict_merge recurses down into dicts nested
    to an arbitrary depth, updating keys. The ``merge_dct`` is merged into
    ``dct``.

    .. versionupdated:: 1.5 - return dct

    :param dct: dict onto which the merge is executed
    :param merged_dict_list: dct merged into dct
    :return: dct
    """
    for dict_item in merged_dict_list:
        for k, v in dict_item.items():  # 2022-04-22 dict_item.iteritems-->dict_item.items
            if (k in dct and isinstance(dct[k], dict)
                    and isinstance(dict_item[k], Mapping)):
                merge_dict(dct[k], dict_item[k])
            else:
                dct[k] = dict_item[k]
    return dct  # @2023-04-12 add result


def get_dict_value_by_type(obj, key, value_type, default=None, by_instance=False):
    """
    Get the dict value of the specified key, if value is not the instance of the specified type, return default value.

    .. versionadded:: 1.2

    :param obj:
    :param key:
    :param value_type:
    :param default:
    :param by_instance:
    :return:
    """
    if key not in obj:
        return default
    value = obj.get(key)
    if by_instance is True:
        if value_type is int and (value is True or value is False):  # instance(True,input) == True
            return default
        if isinstance(value, value_type):
            return value
    else:
        if type(value) is value_type:
            return value
    return default


# -------------------------------------------type-------------------------------------------
def is_str(value, by_instance=False):
    """
    Check whether the value is string.

    :param value:
    :param by_instance:
    :return:
    """
    if by_instance is True:
        return isinstance(value, str)
    return type(value) == str


def is_list(value, by_instance=False):
    """
    Check whether the value is list object.

    :param value:
    :param by_instance:
    :return:
    """
    if by_instance is True:
        return isinstance(value, list)
    return type(value) == list


def is_dict(value, by_instance=False):
    """
    Check whether the value is dict object.

    :param value:
    :param by_instance:
    :return:
    """
    if by_instance is True:
        return isinstance(value, dict)
    return type(value) == dict


# -------------------------------------------str-------------------------------------------
def get_wrap_str(*items):
    """
    Use '\n' to join multiple strings.

    :param items:
    :return:
    """
    result = []
    for item in items:
        if item is not None:
            result.append(str(item))
    return '\n'.join(result)


def slice_str(value, start_len=6, end_len=0, ellipsis_str='......'):
    if end_len != 0:
        end_len = -abs(end_len)

    value = str(value)
    if len(value) <= (start_len - end_len):
        return value
    result = value[0:start_len]
    if end_len != 0:
        result = result + ellipsis_str + value[end_len:]
    return result


def str_replace(txt, old, new=''):
    """
    Return a copy with all occurrences of substring old replaced by new.

    .. versionadded:: 1.5

    Example:
        str_replace('abca','a','A')             # AbcA
        str_replace('abca',{'a':'A','b':'B'})   # ABcA
        str_replace('abca',['a','b'],"*")       # **c*

    :param txt:
    :param old:
    :param new:
    :return:
    """
    v_type = type(old)
    if v_type is dict:
        for key in old:
            txt = txt.replace(key, old[key])
    elif v_type is list:
        for item in old:
            txt = txt.replace(item, new)
    else:
        txt = txt.replace(old, new)
    return txt


# -------------------------------------------other-------------------------------------------
def bulk_append_child(items, parent_map, item_parent_key='parent_id', children_key="children", parent_map_key='id'):
    """
    Append the items to the child list of the matched parent object.

    Example:
        bulk_append_child(interface_items,device_list, 'device_id', 'interface_list')

    :param items:
    :param item_parent_key:
    :param parent_map:
    :param children_key:
    :param parent_map_key:
    :return:
    """
    if not isinstance(items, list):
        items = [items]

    if isinstance(parent_map, list):
        parent_map = get_dict_mapping(parent_map, parent_map_key)

    for item in items:
        parent_attr = item.get(item_parent_key)
        parent_item = parent_map.get(parent_attr)
        if isinstance(parent_item, dict):
            p_children = parent_item[children_key] = parent_item.get(children_key) or []
            p_children.append(item)


def parse_version(version):
    """
    Parse version string to int list, if version is integer, it will be converted to int

    Example:
        parse_version(sqlalchemy.__version__)   # --> [2, 0, 20] / [2, 0, '0rc1']

    :param version: the version to be parsed, str/int/float/package
    """
    version_type = type(version)

    if version_type is int:  # 1-->1
        return [version]

    if version_type is float:  # 1.2-->[1,2]
        version = str(version)
    elif version_type is not str:  # sqlalchemy,sys
        version = getattr(version, '__version__', None) or getattr(version, 'version', None)

    if type(version) is not str:
        return []

    versions = []
    for v in version.split('.'):
        if v.isdigit():
            versions.append(int(v))
        else:
            versions.append(v)
    return versions



from flask import current_app

from ._app import get_app_config

__all__ = ['create_response', 'get_status_msg', 'ResponseManager']


def create_response(success, data, data_wrapped=False):
    """
    Create the response json result.

    :param success:
    :param data:
    :param data_wrapped:
    :return:
    """
    if success is True:
        return _create_success_response(data, data_wrapped)
    else:
        return _create_fail_response(data)


def _create_success_response(data, data_wrapped=False):
    """
    :param data:
    :param data_wrapped:
    :return:
    """
    status = get_app_config('FLASKZ_RES_SUCCESS_STATUS') or 'success'
    if data_wrapped is True:
        _data = {
            'status': status,
        }
        _data.update(data)
        return _data
    else:
        return {
            'status': status,
            'data': data
        }


def _create_fail_response(status_code, msg=None):
    """
    :param msg:
    :param status_code:
    :return:
    """
    status = get_app_config('FLASKZ_RES_FAIL_STATUS') or 'fail'
    msg = msg or get_status_msg(status_code)

    if type(status_code) == tuple:
        status_code = status_code[0]

    return {
        'status': status,
        'status_code': status_code,
        'message': str(msg),
    }


def get_status_msg(status_code):
    """
    Get the specified message by status_code.
    Can be used to return internationalized text, Local can be fixed, or get the local from request.

    :param status_code:
    :return:
    """
    response_callback = get_current_response_manager('get_response_callback')
    if response_callback:
        return response_callback(status_code)

    if type(status_code) == tuple:
        len_ = len(status_code)
        if len_ > 1:
            return status_code[1] or status_code[0]
        elif len_ > 0:
            return status_code[0]
    return status_code


def get_current_response_manager(callback_name):
    if current_app:
        response_manager = getattr(current_app, 'response_manager', None)
        if response_manager:
            return getattr(response_manager, callback_name, None)


class ResponseManager:
    """
    Used to generate response.
    """

    def __init__(self):
        self._get_response = None

    def init_app(self, app):
        app.response_manager = self

    def get_response(self, get_response):
        self._get_response = get_response

    @property
    def get_response_callback(self):
        return self._get_response



"""
pip install paramiko
"""
import errno
import os
import re
import socket
import stat
import time
from contextlib import contextmanager

import paramiko
from paramiko.ssh_exception import SSHException


@contextmanager
def ssh_session(hostname, username, password=None, port=22, **kwargs):
    """
    SSH contextmanager.

    Example:
        with ssh_session(host, username, password, timeout=20) as ssh:
            ssh.run_command('ls -l')

        with ssh_session(host, username, password, timeout=20) as ssh:
            ssh.run_command_list(['show version', 'show running-config'])

        with ssh_session(host, username, password, timeout=20, secondary_password=enable_pwd, recv_endswith=['# ', '$ ', ': ', '? ', '#']) as ssh:
            ssh.run_command_list(['enable', 'show run'])

    """
    ssh_client = SSH(hostname=hostname, username=username, password=password, port=port, **kwargs)
    yield ssh_client
    ssh_client.close()


class SSH(object):
    def __init__(self, hostname, username, password=None, port=22, **kwargs):
        """
        Create a SSH instance.

        .. versionupdated:: 1.6 - add secondary_password and recv_endswith kwargs

        Example:
            ssh = SSH(host, username, password, timeout=20)
            ssh.run_command('ls -l')

            ssh = SSH(host, username, password, timeout=20)
            ssh.run_command_list(['enable', enable_pwd, 'show run'])

            ssh = SSH(host, username, password, timeout=20, secondary_password=enable_pwd, recv_endswith=['# ', '$ ', ': ', '? ', '#'])
            ssh.run_command_list(['enable', 'show run'])


        :param hostname: the host(address) to ssh
        :param username: the username of the host
        :param password: the password of the host
        :param port: the ssh port, default is 22
        :param kwargs:
                - secondary_password: use for enable/sudo action, if None, the password needs to be sent through the command
                - recv_endswith: use for stop receiving, default is ['# ', '$ ', ': ', '? ']
                - timeout: timeout on blocking read/write operations & socket timer, default is 0
                - connect_kwargs: kwargs for Transport.connect()
        """
        self._username = username
        self._password = password
        self._secondary_password = kwargs.pop('secondary_password', None)  # for enable/sudo
        # recv_endswith = kwargs.pop('recv_endswith', None)  # for stop receiving
        self.recv_endswith = tuple(kwargs.pop('recv_endswith', None) or ['# ', '$ ', ': ', '? '])  # for stop receiving
        self._timeout = kwargs.pop('timeout', 0)

        # self.transport = paramiko.Transport((hostname, port))
        self.transport = paramiko.Transport(_create_socket(hostname=hostname, port=port, timeout=self._timeout))
        _connect_kwargs = kwargs.pop('connect_kwargs', None) or {}  # kwargs for Transport.connect()
        self.transport.connect(username=username, password=password, **_connect_kwargs)

        self.ssh = paramiko.SSHClient()
        self.ssh._transport = self.transport
        self.ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())

    @property
    def sftp(self):
        if not hasattr(self, '_sftp'):
            self._sftp = paramiko.SFTPClient.from_transport(self.transport)
        return self._sftp

    @property
    def channel(self):
        if not hasattr(self, '_channel'):
            self._channel = self.ssh.invoke_shell(height=100000)
            if self._timeout > 0:
                self._channel.settimeout(self._timeout)
        return self._channel

    def run_command(self, command, recv=True, clean=True):
        """
        Run the command.
        If recv is False, just run the command and return immediately, without waiting for the result

        .. versionupdated::
            1.6.1   - @2023-06-26 add recv parameter
            1.6.3   - @2023-07-10 add clean parameter

        Example:
            ssh.run_command('ls -l')
            ssh.run_command('show run')
            ssh.run_command({
                'command': ' ',
                'clean': False,
            })

         :param command: the command to run.
         :param recv: wait for the result or not.
         :param clean: clean output info or not, default True(clean)

         :return: the output of the command
        """
        _command, _recv, _clean = _get_command_arg(command, recv=recv, clean=clean)

        _command = _command.strip()
        self.channel.send(_command + '\n')
        if _recv is False:
            return None

        output = self._get_output(_command, clean=_clean)
        enable_commands = ('sudo', 'enable')
        secondary_password = self._secondary_password  # or self._password
        if secondary_password and \
                _command.lower().startswith(enable_commands) and \
                'assword' in output:  # input password
            pwd_output = self.run_command(secondary_password)
            # if 'assword' in pwd_output:
            if pwd_output == output:
                return output
            return self.run_command(_command)

        return output

    def run_command_list(self, command_list, last_result=False):
        """
        Run a command list.
        By default, returns the list of results , if last_result==True returns the result of the last command

        Example:
            ssh.run_command_list(['cd /usr/projects/git/srte',
                      'pwd',
                      'git pull origin master',
                      'wiui@hotmail.com',
                      '11111111'
                      ], True)

         :param command_list: the command list to run.
         :param last_result: if True, return the result of the last command, otherwise, return the result list

         :return: command output result(list/last)
        """
        re_list = []
        for command in command_list:
            re_list.append(self.run_command(command))
        if last_result is True:
            return re_list[-1]
        return re_list

    def sftp_get_dir(self, remote_dir, local_dir):
        """
        Download remote direction to local.
        Return the local files list.

        Example:
            ssh.sftp_get_dir('/usr/projects/git/srte/src/', '/Users/taozh/Work/Codes/ssh_test/sftp')
        """
        if not self._path_exists(remote_dir):
            return False
        remote_dir = _remove_end_slash(remote_dir)
        local_dir = _remove_end_slash(local_dir)
        if not os.path.exists(local_dir):
            os.mkdir(local_dir)
        all_files = self.listdir(remote_dir, True)
        local_files = []
        for f in all_files:
            local_filename = f.replace(remote_dir, local_dir)
            local_filepath = os.path.dirname(local_filename)
            local_files.append(local_filepath)
            if not os.path.exists(local_filepath):
                os.makedirs(local_filepath)
            self.sftp.get(f, local_filename)
        return local_files

    def listdir(self, path, recursion=False):
        """
        List all files in the given path.

        Example:
            ssh.listdir('/usr/projects/git/srte/src/')
        """
        all_files = []
        path = _remove_end_slash(path)
        if path[-1] == '/':
            path = path[0:-1]
        files = self.sftp.listdir_attr(path)
        for f in files:
            filename = path + '/' + f.filename
            if stat.S_ISDIR(f.st_mode):  # 如果是文件夹的话递归处理
                if recursion is True:
                    all_files.extend(self.listdir(filename, recursion))
            else:
                all_files.append(filename)
        return all_files

    def close(self):
        """Close the connect"""
        self.channel.close()
        self.ssh.close()

    def _path_exists(self, path):
        """Return whether the path exists"""
        try:
            self.sftp.stat(path)
        except IOError as e:
            if e.errno == errno.ENOENT:
                return False
            raise
        else:
            return True

    def _get_output(self, command, clean=True):
        output = self._recv_data()
        if clean is not False:
            output = _clean_output_info(output, command)
        return output

    def _recv_data(self):
        """Receive the command output"""
        while not self.channel.recv_ready():
            time.sleep(0.01)
        res_list = []
        time.sleep(0.2)  # Solve the problem of incomplete data
        # cmd_pattern = re.compile('.*[#$] ' + command) # Does not work with password entry

        while True:
            data = self.channel.recv(1024)
            info = data.decode()
            res = info.replace(' \r', '')
            res_list.append(res)
            if len(info) < 1024:  # read speed > write speed
                if info.endswith(self.recv_endswith):  # Cisco C9300:# /
                    break

        return ''.join(res_list)


def _get_command_arg(command, recv, clean):
    if type(command) is dict:
        _command = command.get('command')
        _recv = command.get('recv') is not False if 'recv' in command else recv
        _clean = command.get('clean') is not False if 'clean' in command else clean
    else:
        _command = command
        _recv = recv
        _clean = clean
    return _command, _recv, _clean


def _create_socket(hostname, port, timeout=0):
    """
    Create socket instance.
    If timeout>0, set the timeout of the instance, otherwise use the default timeout(maybe 75s).

    :param hostname:
    :param port:
    :param timeout:
    :return:
    """

    reason = 'No suitable address family'
    addrinfos = socket.getaddrinfo(hostname, port, socket.AF_UNSPEC, socket.SOCK_STREAM)
    for family, socktype, proto, canonname, sockaddr in addrinfos:
        if socktype == socket.SOCK_STREAM:
            sock = socket.socket(family, socket.SOCK_STREAM)
            if timeout > 0:
                sock.settimeout(timeout)
            try:
                # retry_on_signal(lambda: sock.connect((hostname, port))) # @2023-06-09 change, remove paramiko.util.retry_on_signal(Paramiko3.0.0 2023-01-20)
                sock.connect((hostname, port))
            except socket.error as e:
                reason = str(e)
            else:
                break
    else:
        raise SSHException(
            'Unable to connect to {}: {}'.format(hostname, reason)
        )
    return sock


def _clean_output_info(txt, command):
    """
    Clear the redundant information.
    - Welcome info      ex)Welcome to Ubuntu...
    - Last login info   ex)Last login...
    - Path info         ex)[root@localhost ~]...
    - Command info      ex)ls -l

    """
    # remove the start command  ex) ls -l
    if txt.startswith(command):
        txt = txt[len(command):].lstrip()

    # remove text before command ex) [root@localhost ~]# ls -l
    cmd_pattern = re.compile('.*([#$])( )*' + re.escape(command))  # @2023-06-07 change ([#$])? --> ([#$])
    last_match = None
    for match in cmd_pattern.finditer(txt):
        last_match = match
    if last_match:
        txt = txt[last_match.end():]
    path_pattern = re.compile('.*[#$]( )?')  # @2023-07-06 change '.*[#$] ' --> '.*[#$]( )?'
    txt = path_pattern.sub('', txt)  # remove the path info
    txt = txt.replace(command + '\r\n', '')  # remove the command
    return txt.strip()


def _remove_end_slash(path):
    """Remove ending slash """
    if path[-1] == '/':
        return path[0:-1]
    return path



from sqlalchemy import Integer, Numeric

from .. import res_status_codes

__all__ = ['BaseModelMixin']


class BaseModelMixin:
    # -------------------------------------------about model-------------------------------------------
    @classmethod
    def get_class_name(cls):
        """
        Get the name of the model class.

        Example:
            User.get_class_name() #User

        :return:
        """
        return cls.__name__

    @classmethod
    def get_columns(cls):
        """
        Get all the columns of the model class.

        Example:
            User.get_cls_columns()

        :return:
        """
        return cls.__table__.columns

    @classmethod
    def get_columns_fields(cls):
        """
        Get all the column fields.

         .. versionadded:: 1.3

        Example:
            User.get_columns_fields()

        :return:
        """
        fields = []
        for col in cls.get_columns():
            fields.append(cls.get_column_field(col))  # col.key
        return fields

    @classmethod
    def get_column_field(cls, col):
        """
        Get the field variable name of the column, default return the key of the column.
        Used to get json fields to create or update data(filter_attrs_by_columns), or convert data to json(_to_json).

        If the field name is different from the database column name
        #1 set field in info dict.(recommend)
        system_default = Column('default', Boolean, default=False, info={'field': 'system_default'})

        #2 rewrite to get the field name
        system_default = Column('default', Boolean, default=False)
        @classmethod
        def get_column_field(cls, col):
            key = col.key
            if key == 'default':
                return 'system_default'
            return key

        :param col:
        :return:
        """
        field = col.info.get('field')
        if field:
            return field
        return col.key

    @classmethod
    def get_column_by_field(cls, field):
        """
        Get the column of the specified field.

        Example:
            User.get_column_by_field('name')

        :param field:
        :return:
        """
        for col in cls.get_columns():
            if (cls.get_column_field(col)) == field:
                return col

    @classmethod
    def get_primary_column(cls):
        """
        Get the primary column of the model class.

        Example:
            User.get_primary_column()   # id column

        :return:
        """
        return find_list(cls.get_columns(), lambda c: c.primary_key is True)

    @classmethod
    def get_primary_key(cls):
        """
        Get the key of the primary column.

        Example:
            User.get_primary_key()  # 'id'

        :return:
        """
        col = cls.get_primary_column()
        if col is not None:  # must
            return col.key
        return None

    @classmethod
    def get_primary_field(cls):
        """
        Get the field of the primary column.

        Example:
            User.get_primary_field()    # 'id'

        :return:
        """
        col = cls.get_primary_column()
        if col is not None:  # must
            return cls.get_column_field(col)  # col.key
        return None

    @classmethod
    def get_unique_columns(cls):
        """
        Get the unique column list of the model class.

        Example:
            User.get_unique_columns()

        :return:
        """
        return filter_list(cls.get_columns(), lambda col: col.unique is True)

    @classmethod
    def get_relationships(cls):
        """
        Get all the relationship list of the model class.

        Example:
            User.get_relationships()

        :return:
        """
        return cls.__mapper__.relationships

    # -------------------------------------------About data-------------------------------------------
    def refresh(self):
        """
        Expire and refresh attributes on the current instance.

        .. versionadded:: 1.6.4

        Example:
            user = User.query_by({'name': 'flaskz'}, True)
            # ...
            user.refresh()

        :return:
        """
        _refresh_instance(self)  # @2023-10-17 add

    def to_dict(self, option=None):
        """
        Convert model data to dict.

        ins_to_dict(A, {
            'cascade': 3,  # 如果子项没有cascade，则使用父项-1，如果cascade不大于0，则不对象属性
            'recursion_value': '{...}'
            # 'include': ['a1'],  # 只有include中的字段才会返回，不区分值是否是对象，include的优先级>exclude
            # 'exclude': ['a2'],  # 只有exclude外的字段才会返回，不区分值是否是对象
            # 'getattr_items': lambda ins, item_cascade, option, path_keys: ins.__dict__.items(),
            #
            # 'bb': {  # 某个子项的设置
            #     # 'cascade': 1,
            #     # 'include': ['b1'],
            #     'exclude': ['b2'],
            # },
            # 'bb.xx': {  # 子项的子项的设置
            #     'exclude': ['x1']
            # },
            # 'bb.yy': {
            #     'exclude': ['y2']
            # },
            # 'cc': {  # cc列表中的每个数据的设置
            #     'exclude': ['n'],
            #     'cascade': 1
            # },
            # "cc.xx": {
            #     'exclude': ['x2']
            # }
        })

        过滤
            -不同的class会有不同的条件过滤-->过滤掉不需要的属性(create_user/password)
            -过滤条件可以是class上的属性，也可以是传递到方法中的exclude
                -如果有include--->只返回include中的属性
                -如果没有include而有exclude--->将exclude中的属性过滤掉
        # 如果属性是关系对象/列表，则



        # 单个对象很容易处理，但是 relationships不好处理
        # 不同的relationship，cascade不同
        # relation中可能有嵌套
        # 有可能两个对象里都有同一个对象
        """
        opt = {
            'getattrs': lambda ins, *args, **kwargs: ins.__class__.get_to_dict_attrs(ins, *args, **kwargs),
            'include': lambda ins, key: ins.__class__.to_dict_field_filter(key),
        }
        if option:
            opt.update(option)
        return ins_to_dict(self, opt)

    @classmethod
    def get_to_dict_attrs(cls, ins, cascade, *args):
        """to_dict attr """
        fields = [cls.get_column_field(col) for col in cls.get_columns()]
        if cascade > 0:
            for relationship in cls.get_relationships():
                lazy = relationship.lazy
                if lazy != 'dynamic' and lazy != 'noload':
                    fields.append(relationship.key)
        items = {}
        for field in fields:
            items[field] = getattr(ins, field, None)
        return items

    @classmethod
    def to_dict_field_filter(cls, field):
        """
        to_dict field filter callback.
        If return false, the field will not be returned.
        :param field:
        :return:
        """
        return True

    @classmethod
    def filter_attrs_by_columns(cls, data):
        """
        Filter out the attributes(dict) corresponding to the columns from the specified data(dict).

        :param data:
        :return:
        """
        attrs = {}
        auto_columns = getattr(cls, 'auto_columns', [])
        auto_fields = []
        for col in auto_columns:
            if is_str(col):
                auto_fields.append(col)
            else:
                auto_fields.append(cls.get_column_field(col))

        for col in cls.get_columns():
            field = cls.get_column_field(col)  # col.key
            if (field in data) and (field not in auto_fields) and (not col.info.get('auto', False)):
                col_type = col.type
                value = data.get(field)
                is_blank_str = is_str(value) and value.strip() == ""
                if col.nullable is False:
                    if not (value is None or is_blank_str):
                        attrs[field] = value
                else:
                    # @2023-05-16 add to fix DataError: (1366, "Incorrect integer value: '' for column")
                    if is_blank_str and (isinstance(col_type, Integer) or isinstance(col_type, Numeric)):
                        attrs[field] = None
                    else:
                        attrs[field] = value

        return attrs

    @classmethod
    def get_columns_json(cls, data):
        """
        Get the attributes(dict) corresponding to the column from the specified data(dict).

        .. deprecated:: 1.5 please use filter_attrs_by_columns method

        :param data:
        :return:
        """

        return cls.filter_attrs_by_columns(data)

    # -------------------------------------------add-------------------------------------------
    @classmethod
    def check_add_data(cls, data):
        """
        Check the the added json data.
        --validate the json data
        --check unique
        If the check result is not True, the adding process will be terminated and the check result will be returned to the client.

        :param data:  The data to be added
        :return: True|Error Message
        """
        if not is_dict(data):
            if data:
                return data
            return res_status_codes.bad_request
        return cls._check_unique(data)

    @classmethod
    def add_db(cls, data):
        """
        Add data to the db.

        Example:
            ins = User.add_db({"name": "taozh", "email": "taozh@focus-ui.com"})

        :param data:
        :return:
        """

        instance = create_instance(cls, data)
        with _db_session(False, True) as session:
            session.add(instance)
            session.commit()
            session.refresh(instance)  # If not, the value in the instance is not consistent with the database, and the query will be send until accessing the attributes.

        return instance

    @classmethod
    def bulk_add(cls, items, with_relationship=False):
        """
        Perform a bulk add of the given list of mapping dictionaries.(atomic)

        sa version upgrade: https://docs.sqlalchemy.org/en/20/orm/queryguide/dml.html
        :param items:
        :param with_relationship:
        :return:
        """
        if len(items) == 0:
            return
        if with_relationship is True:
            with db_session() as session:
                ins_list = []
                for data_item in items:
                    ins_list.append(create_instance(cls, data_item))
                session.add_all(ins_list)
        else:
            with db_session() as session:
                session.bulk_insert_mappings(cls, items)

    # -------------------------------------------update-------------------------------------------
    @classmethod
    def check_update_data(cls, data):
        """
        Check the the updated json data.
        --validate the json data
        --check exist
        --check unique
        If the check result is not True, the update process will be terminated and the check result will be returned to the client.

        :param data: The updated json data
        :return:
        """
        if not is_dict(data):
            if data:
                return data
            return res_status_codes.bad_request

        exist = cls._check_exist(data)
        if exist is not True:
            return exist
        return cls._check_unique(data)

    @classmethod
    def update_db(cls, data):
        """
        Update the data to the db. The primary key value must be in data.
        Only the fields in data will be updated.

        Example:
            ins = User.update_db({"id": 1, "email": "taozh@focus-ui.com"})

        :param data:
        :return:
        """
        pk_value = cls._get_pk_value(data)
        if pk_value is None:  # pk value does not exist
            return res_status_codes.db_data_not_found

        with _db_session(False, True) as session:
            instance = _session_get(session, cls, pk_value)  # session.query(cls).get(pk_value)  # Scenarios for extending BaseModelMixin instead of ModelMixin
            if instance is None:  # Object does not exist
                return res_status_codes.db_data_not_found
            cls._update_ins(instance, data)  # @2022-12-01 change to ensure the updated instance and the setattr action in the same session
            session.commit()
            session.refresh(instance)  # If not, the value in the instance is not consistent with the database, and the query will be send until accessing the attributes.
        return instance

    @classmethod
    def bulk_update(cls, items, with_relationship=False):
        """
        Perform a bulk update of the given list of mapping dictionaries.(atomic)

        sa version upgrade: https://docs.sqlalchemy.org/en/20/orm/queryguide/dml.html
        :param items:
        :param with_relationship:
        :return:
        """
        if len(items) == 0:
            return

        if with_relationship is True:
            with db_session() as session:
                for data in items:
                    pk_value = cls._get_pk_value(data)
                    if pk_value is not None:
                        # cls._update_ins(session.query(cls).get(pk_value), data)  # @2022-12-01 change to ensure the updated instance and the setattr action in the same session
                        cls._update_ins(_session_get(session, cls, pk_value), data)  # @2022-12-01 change to ensure the updated instance and the setattr action in the same session
        else:
            with db_session() as session:
                session.bulk_update_mappings(cls, items)

    @classmethod
    def _update_ins(cls, instance, data):
        if instance:
            for field, value in cls.filter_attrs_by_columns(data).items():  # Only the fields in json will be updated
                setattr(instance, field, value)  # @2023-05-11, data.get(field)-->ins_attrs.get(field)
            relationships = create_relationships(cls, data)
            for field in relationships:
                setattr(instance, field, relationships[field])
        return instance

    # -------------------------------------------delete-------------------------------------------
    @classmethod
    def check_delete_data(cls, pk_value):
        """
        Check the the deleted data.
        --validate the deleted data
        --check exist
        If the check result is not True, the delete process will be terminated and the check result will be returned to the client.

        :param pk_value:
        :return:
        """
        if pk_value is None:
            return res_status_codes.db_data_not_found

        pk = cls.get_primary_field()
        data = {pk: pk_value}
        return cls._check_exist(data)

    @classmethod
    def delete_db(cls, pk_value):
        """
        Delete the specified data with the specified primary key value.

        Example:
            ins = User.delete_db(1)   # pk
            ins = User.delete_db({'id': 1})   # dict

        :param pk_value:
        :return:
        """
        if pk_value is None:  # Object does not exist
            return res_status_codes.db_data_not_found
        with db_session() as session:  # @2022-12-01 change to ensure the deleted instance and the delete action in the same session
            if is_dict(pk_value):  # @2023-03-27 add
                instance = session.query(cls).filter_by(**pk_value).limit(1).first()
            else:
                # instance = session.query(cls).get(pk_value)
                instance = _session_get(session, cls, pk_value)
            if instance is None:  # Object does not exist
                return res_status_codes.db_data_not_found
            session.delete(instance)
        return instance

    @classmethod
    def bulk_delete(cls, items):
        """
        Perform a bulk delete of the given list of mapping dictionaries.(not atomic)

        .. versionupdated:: 1.0

        Example:
            User.bulk_delete([1,2,3])   # pk list
            User.bulk_delete([{'id': 1}, {'id': 2}, {'id': 3}]) # dict list

        :param items:
        :return: the deleted count
        """
        if len(items) == 0:
            return
        pk = cls.get_primary_field()
        count = 0
        with db_session() as session:
            pk_list = []
            for item in items:
                if is_dict(item):
                    # instance = session.query(cls).filter_by(**item).limit(1).first()
                    # if instance:
                    #     pk_list.append(getattr(instance, pk))
                    for instance in session.query(cls).filter_by(**item):  # @2023-03-27 first to all, ex)delete all items with same name
                        pk_list.append(getattr(instance, pk))
                else:
                    pk_list.append(item)
            # @2022-11-03: Use 'where' and 'in' to rewrite bulk delete to avoid partial delete scenarios
            if len(pk_list) > 0:
                count = session.query(cls).filter(getattr(cls, pk).in_(pk_list)).delete()
        return count

    @classmethod
    def clear_db(cls):
        """
        Clear all the data.

        .. versionadded:: 1.6

        Example:
            SysActionLog.clear_db()

        :return: the deleted count
        """
        with db_session() as session:
            return session.query(cls).delete()

    # -------------------------------------------query-------------------------------------------

    @classmethod
    def query(cls):
        """
        Return a 'Query' object corresponding to this class.

        Example:
            query = TemplateModel.query()
            print(query.all())

        :return:
        """
        with db_session(do_commit=False) as session:
            return session.query(cls)

    @classmethod
    def get_query_default_order(cls):
        """
        Get the default order of the query.(column key, not field)

        get_primary_key-->get_primary_column to fix
        sqlalchemy.exc.CompileError: Can't resolve label reference for ORDER BY / GROUP BY / DISTINCT etc.
            Textual SQL expression 'id' should be explicitly declared as text('id')
        :return:
        """
        return cls.get_primary_column()

    @classmethod
    def query_by_unique_key(cls, data):
        """
        Query data by the unique values.
        --If exist,returns the instance.
        --Otherwise,returns None

        Example:
            ins = User.query_by_unique_key(data)

        :param data:
        :return:
        """
        cols = cls.get_unique_columns()
        if len(cols) == 0:
            return None

        ors = []
        for col in cols:
            field = cls.get_column_field(col)  # col.key
            value = data.get(field)
            if value is not None:  # maybe 0
                ors.append(get_col_op(col, '==', value))  # @2023-08-16, text-->op

        if len(ors) == 0:
            return None

        with db_session(do_commit=False) as session:
            query = session.query(cls)
            query = append_query_filter(query, ors, 'or')
            instance = query.first()
        return instance

    @classmethod
    def query_by_pk(cls, pk_value):
        """
        Query by pk value.

        Example:
            ins = User.query_by_pk(1)

        :param pk_value:
        :return:
        """
        with db_session(do_commit=False) as session:
            # instance = session.query(cls).get(pk_value)
            instance = _session_get(session, cls, pk_value)
        return instance

    @classmethod
    def query_by(cls, by_dict, return_first=False):
        """
        Query by dict object.
        -If first is not True, return the list result of the query.
        -If first is True, return the first row object or None.

        Example:
            ins_list = User.query_by({'name': 'flaskz'})   # list
            ins = User.query_by({'name': 'flaskz'}, True) # first row

        """
        with db_session(do_commit=False) as session:
            if return_first is True:
                result = session.query(cls).filter_by(**by_dict).limit(1).first()
            else:
                result = session.query(cls).filter_by(**by_dict).all()
        return result

    @classmethod
    def query_all(cls):
        """
        Query all the data of the model class.

        Example:
            ins_list = User.query_all()

        :return:
        """
        query_order = cls.get_query_default_order()
        with db_session(do_commit=False) as session:
            query = session.query(cls)
            if query_order is not None:
                query = query.order_by(query_order)
            result = query.all()
        return result

    @classmethod
    def query_pss(cls, pss_option):
        """
        Query data by search, pagination and sort condition.
        Please use flaskz.utils.get_pss to parse option first.
        pss = page+search+sort

        Example:
            result = TemplateModel.query_pss(get_pss(   # use flaskz.utils.get_pss to format condition
                TemplateModel, {   # FROM templates
                    "search": {                         # WHERE
                        "like": "t",                    # name like '%t%' OR description like '%t%' (TemplateModel.like_columns = ['name', description])
                        "age": {                        # AND (age>1 AND age<20)
                            ">": 1,                     # operator:value, operators)'='/'>'/'<'/'>='/'<='/'BETWEEN'/'LIKE'/'IN'
                            "<": 20
                        },
                        "email": "taozh@focus-ui.com",  # AND (email='taozh@focus-ui.com')
                        "_ors": {                       # AND (country='America' OR country='Canada')
                            "country": "America||Canada"
                        },
                        "_ands": {                      # AND (grade>1 AND grade<5)
                            "grade": {
                                ">": 1,
                                "<": 5
                            }
                        }
                    },
                    "sort": {                           # ORDER BY templates.name ASC
                        "field": "name",
                        "order": "asc"
                    },
                    # "sort":[                          # ORDER BY templates.name ASC, templates.age DESC
                    #     {"field": "name", "order": "asc"},
                    #     {field": "age", "order": "desc"}
                    # ],
                    "page": {                           # LIMIT ? OFFSET ? (20, 0)
                        "offset": 0,
                        "size": 20
                    }
                }))

        # sql
        SELECT templates.id AS templates_id, templates.name AS templates_name, templates.age AS templates_age, templates.email AS templates_email,
                templates.country AS templates_country, templates.grade AS templates_grade, templates.description AS templates_description,
                templates.created_at AS templates_created_at, templates.updated_at AS templates_updated_at
        FROM templates
        WHERE
            (name like '%t%' OR description like '%t%')
            AND (grade>1 AND grade<5 AND age>1 AND age<20 AND email='taozh@focus-ui.com')
            AND (country='America' OR country='Canada')
        ORDER BY templates.name ASC
        LIMIT ? OFFSET ? (20, 0)

        :param pss_option:
        :return:
        """
        return cls._query_pss(pss_option)

    @classmethod
    def count(cls, search=None):
        """
        Return the count of the specified search, if search option is None, return the count of all data.

        .. versionadded:: 1.6

        Example:
            SysActionLog.count()
            SysActionLog.count(get_pss(   # use flaskz.utils.get_pss to format condition
                TemplateModel, {   # FROM templates
                    "search": {                         # WHERE
                        "like": "t",                    # name like '%t%' OR description like '%t%' (TemplateModel.like_columns = ['name', description])
                        "age": {                        # AND (age>1 AND age<20)
                            ">": 1,                     # operator:value, operators)'='/'>'/'<'/'>='/'<='/'BETWEEN'/'LIKE'/'IN'
                            "<": 20
                        },
                        "email": "taozh@focus-ui.com",  # AND (email='taozh@focus-ui.com')
                        "_ors": {                       # AND (country='America' OR country='Canada')
                            "country": "America||Canada"
                        },
                        "_ands": {                      # AND (grade>1 AND grade<5)
                            "grade": {
                                ">": 1,
                                "<": 5
                            }
                        }
                    }
                }))
        :param search: the search option
        :return: the count number
        """
        return cls._query_pss(search, True)

    @classmethod
    def _query_pss(cls, pss_option, return_count=False):
        pss_option = pss_option or {}
        filter_likes = pss_option.get('filter_likes', [])
        filter_ands = pss_option.get('filter_ands', [])
        filter_ors = pss_option.get('filter_ors', [])

        group = pss_option.get('group', [])

        offset = max(get_dict_value_by_type(pss_option, 'offset', int, 0), 0)  # @2023-01-09 update, add type check
        limit = max(get_dict_value_by_type(pss_option, 'limit', int, 0), 0)
        # distinct = pss_option.get('distinct', [])

        orders = pss_option.get('order')
        if orders is None:
            orders = []
        elif not is_list(orders):  # asc/desc
            orders = [orders]

        if len(orders) == 0:  # @2023-05-04 fix
            orders = [cls.get_query_default_order()]  # default order
        orders = [item for item in orders if item is not None]

        with db_session(do_commit=False) as session:
            query = session.query(cls)

            query = append_query_filter(query, filter_likes, 'or')
            query = append_query_filter(query, filter_ands, 'and')
            query = append_query_filter(query, filter_ors, 'or')
            # if len(filter_likes) > 0:
            #     query = query.filter(text('(' + (' OR '.join(filter_likes)) + ')'))
            # if len(filter_ands) > 0:
            #     query = query.filter(text('(' + (' AND '.join(filter_ands)) + ')'))
            # if len(filter_ors) > 0:
            #     query = query.filter(text('(' + (' OR '.join(filter_ors)) + ')'))

            # if len(distinct) > 0:
            #     query.distinct(*distinct)

            if len(group) > 0:  # @2023-06-07 add
                query = query.group_by(*group)

            count = query.count()

            if return_count is True:
                return count

            if count > 0 and offset < count:
                # for order in orders:
                #     if order is not None:
                #         query = query.order_by(order)
                if len(orders) > 0:
                    query = query.order_by(*orders)

                query = query.offset(offset)
                if limit > 0:
                    query = query.limit(limit)
                items = query.all()
            else:
                items = []
        return {
            'count': count,
            'data': items
        }

    # -------------------------------------------check-------------------------------------------
    @classmethod
    def _get_pk_value(cls, data):
        """
        Get the primary key value from the json data.

        :param data:
        :return:
        """
        pk = cls.get_primary_field()
        if pk:
            return data.get(pk)

    @classmethod
    def _check_exist(cls, data):
        """
        Check whether the data exists.
        Return True if exists, else return not found code.

        :param data:
        :return:
        """
        pk_value = cls._get_pk_value(data)
        if pk_value is not None and cls.query_by_pk(pk_value):
            return True
        return res_status_codes.db_data_not_found  # used to return to the client

    @classmethod
    def _check_unique(cls, data):
        """
        Check whether the data meets uniqueness constraints.
        If meets, returns True, otherwise returns exist code.
        """
        result = cls.query_by_unique_key(data)

        if result is None:
            return True

        pk = cls.get_primary_field()
        pk_value = cls._get_pk_value(data)
        if pk_value is not None and pk_value == getattr(result, pk):
            return True
        return res_status_codes.db_data_already_exist  # used to return to the client

    def __repr__(self):
        cls = self.__class__
        attrs = []
        for col in cls.get_columns():
            field = cls.get_column_field(col)
            attrs.append(field + '=' + str(getattr(self, field, None)))
        return cls.get_class_name() + '(' + (', '.join(attrs)) + ')'

    # -------------------------------------------new-------------------------------------------


# must
from ..utils._cls import ins_to_dict
from ..utils._common import find_list, filter_list, is_str, is_dict, is_list, get_dict_value_by_type
from ._util import create_instance, create_relationships, db_session, append_query_filter, _db_session, _session_get, _refresh_instance
from ._query_util import get_col_op



from datetime import datetime
from threading import Timer

__all__ = ['set_interval', 'set_timeout', 'run_at']


class _IntervalTimer(object):
    def __init__(self, interval, function, args=None, kwargs=None, immediately=False, daemon=True):
        self._timer = None
        self.interval = interval  # milliseconds
        self.function = function
        self.timer_daemon = daemon
        self.args = args if args is not None else []
        self.kwargs = kwargs if kwargs is not None else {}
        self.is_started = False
        if immediately is True:
            self._run()
        else:
            self.start()

    def _run(self):
        self.is_started = False
        if self.function(*self.args, **self.kwargs) is not False:  # The callback function should do exception catching
            self.start()

    def start(self):
        if not self.is_started:
            self._timer = set_timeout(self.interval, self._run, daemon=self.timer_daemon)
            self.is_started = True

    def cancel(self):
        if self._timer:
            self._timer.cancel()
            self._timer = None
            self.is_started = False

    def stop(self):
        self.cancel()


def set_interval(interval, function, args=None, kwargs=None, immediately=False, daemon=True):
    """
    Execute a function periodically at a specified time interval(milliseconds).
    The result of the function execution is False, then the loop is interrupted.

    .. versionadded:: 1.0

    Example:
        t = set_interval(10, print, ('Hello, World!',))
        # t.cancel()    # Stop the timer

    :param interval: time interval in milliseconds
    :param function: callback function, which should do exception catching*
    :param args:
    :param kwargs:
    :param immediately: If True, the function will be executed immediately, otherwise it will be executed after the interval time
    :param daemon: if true, the timer will be daemon(Daemon threads are abruptly stopped at shutdown)
    :return:
    """
    t = _IntervalTimer(interval, function, args=args, kwargs=kwargs, immediately=immediately, daemon=daemon)
    return t


def set_timeout(interval, function, args=None, kwargs=None, daemon=True):
    """
    Set a timer which executes a function once the timer expires(milliseconds).

     .. versionadded:: 1.0

    Example:
        t = set_timeout(10, print, ('Hello, World!',))
        # t.cancel()    # Stop the timer

    :param interval: delay time in milliseconds before the specified function is executed
    :param function: the function to be executed
    :param args: the args of the function
    :param kwargs: the kwargs of the function
    :param daemon: whether the timer is a daemon thread, default True(Daemon thread stops abruptly when main thread is shut down)
    :return:
    """
    t = Timer(interval / 1000, function, args=args, kwargs=kwargs)
    t.daemon = daemon
    t.start()
    return t


def run_at(at_time, function, args=None, kwargs=None, daemon=True, time_format='%Y-%m-%d %H:%M:%S'):
    """
    Call a function at the specified datetime.

    .. versionadded:: 1.6.1

    Example:
        t1 = run_at('2023-06-28 14:53:00', print, ['run_at'])   # time str
        # t1.cancel()    # Stop the timer

        t2 = run_at(1629811200, print, ['run_at'])  # timestamp
        # t2.cancel()    # Stop the timer

    :param at_time: the specified datetime(str/timestamp)
    :param function: the function to be executed
    :param args: the args of the function
    :param kwargs: the kwargs of the function
    :param daemon: whether the timer is a daemon thread, default True(Daemon thread stops abruptly when main thread is shut down)
    :param time_format: the format of the specified datetime
    :return:
    """
    time_type = type(at_time)
    if time_type is str:
        at_time = datetime.strptime(at_time, time_format)
    elif time_type is int:
        at_time = datetime.fromtimestamp(at_time)
    now = datetime.now()
    timeout_ms = (at_time - now).total_seconds()
    if timeout_ms < 0:
        return False
    elif timeout_ms == 0:
        function(*(args if args is not None else []), **(kwargs if kwargs is not None else {}))
        return

    return set_timeout(timeout_ms * 1000, function, args=args, kwargs=kwargs, daemon=daemon)


if __name__ == '__main__':
    set_timeout(1000, print, ['timeout'], daemon=False)
    run_at('2023-06-28 15:51:00', print, ['timeout_at'], daemon=False)
    # run_at(1629811200, print, ['timeout_at'], daemon=False)



__all__ = ['typed_property', 'init_properties']


def typed_property(name, expected_type):
    """
    Example 1:
        class Person:
            name = typed_property('name', str)
            age = typed_property('age', int)

    Example 2:
        String = partial(typed_property, expected_type=str)
        Integer = partial(typed_property, expected_type=int)
        class Person:
            name = String('name')
            age = Integer('age')

    :param name:
    :param expected_type:
    :return:
    """
    storage_name = '_' + name

    @property
    def prop(self):
        return getattr(self, storage_name)

    @prop.setter
    def prop(self, value):
        if not isinstance(value, expected_type):
            raise TypeError('{} must be a {}'.format(name, expected_type))
        setattr(self, storage_name, value)

    return prop


def init_properties(self, extra_kwargs, *args, **kwargs):
    """
    Example:
        class A:
            _fields = ['a', 'b', 'c']

            def __init__(self, *args, **kwargs):
                init_property(self, False, *args, **kwargs)


        class B:
            _fields = ['a', 'b', 'c']

            def __init__(self, *args, **kwargs):
                init_property(self, True, *args, **kwargs)  # extended keyword arguments


        print(A(1, 2, 3))
        print(A(1, 2, c=3))
        print(A(1, b=2, c=3))
        # print(A(1, 2, 3, e=4))  # TypeError: A got an unexpected keyword argument e
        # print(A(1, 2, 3, 4))  # TypeError: A takes 3 positional arguments but 4 were given
        # print(A(1, 2, 3, a=4))  # TypeError: A got an unexpected keyword argument a

        print(B(1, 2, 3))
        print(B(1, 2, c=3))
        print(B(1, b=2, c=3))
        print(B(1, 2, 3, e=4))
        # print(B(1, 2, 3, 4))  # TypeError: B takes 3 positional arguments but 4 were given
        # print(B(1, 2, 3, a=4))  # TypeError: B got an unexpected keyword argument a
    """
    _fields = getattr(self, '_fields', [])
    filed_len = len(_fields)
    args_len = len(args)
    cls_name = self.__class__.__name__
    if args_len > filed_len:
        raise TypeError(cls_name + ' takes {} positional arguments but {} were given'.format(filed_len, args_len))

    for name, value in zip(_fields, args):
        setattr(self, name, value)

    for name in self._fields[args_len:]:
        setattr(self, name, kwargs.pop(name))

    if extra_kwargs is True:
        extra_args = kwargs.keys() - self._fields
        for name in extra_args:
            setattr(self, name, kwargs.pop(name))

    if kwargs:
        raise TypeError(cls_name + ' got an unexpected keyword argument {}'.format(','.join(kwargs)))

    return self



import sqlite3
import time
from datetime import datetime

from flask import Flask
from sqlalchemy import create_engine, event
from sqlalchemy.engine import Engine
from sqlalchemy.engine import ExceptionContext
from sqlalchemy.exc import OperationalError
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker

DBSession = sessionmaker(autocommit=False)

ModelBase = declarative_base()


def init_model(app):
    """
    Initialize the database.

    Example:
        init_model(flask_app)

    """
    is_app = isinstance(app, Flask)
    app_config = get_app_config_items(app) or {}
    database_uri = app_config.get('FLASKZ_DATABASE_URI')

    # enable sqlite foreign key
    # if database_uri.startswith('sqlite'):
    @event.listens_for(Engine, "connect")
    def set_sqlite_pragma(dbapi_connection, connection_record):
        cursor = dbapi_connection.cursor()
        if isinstance(cursor, sqlite3.Cursor):  # for multiple db connect
            cursor.execute("PRAGMA foreign_keys=ON")
        cursor.close()

    try:
        engine_kwargs = {}
        custom_engine_kwargs = app_config.get('FLASKZ_DATABASE_ENGINE_KWARGS')  # @2023-02-27: add FLASKZ_DATABASE_ENGINE_KWARGS config
        if type(custom_engine_kwargs) is dict:
            engine_kwargs.update(custom_engine_kwargs)

        for engine_key, config_key in {'echo': 'FLASKZ_DATABASE_ECHO',
                                       'pool_recycle': 'FLASKZ_DATABASE_POOL_RECYCLE',
                                       'pool_pre_ping': 'FLASKZ_DATABASE_POOL_PRE_PING'}.items():  # @2023-02-01: add pool_pre_ping config
            if config_key in app_config:
                engine_kwargs[engine_key] = app_config.get(config_key)

        engine = create_engine(database_uri, **engine_kwargs)
        session_kwargs = app_config.get('FLASKZ_DATABASE_SESSION_KWARGS') or {}  # @2023-08-25: add FLASKZ_DATABASE_SESSION_KWARGS config
        DBSession.configure(binds={ModelBase: engine}, **session_kwargs)  # for multiple db

        with engine.connect():  # connect test
            flaskz_logger.info('Database ' + database_uri + ' is ready\n')

        # handle disconnect error
        engine_err_info = {}

        @event.listens_for(engine, "handle_error")
        def handle_engine_error(context: ExceptionContext):  # @2023-02-01: add error handler
            original_exception = getattr(context, 'original_exception', None)
            flaskz_logger.error('Engine error:\n' + str(original_exception))
            if not isinstance(original_exception, OperationalError):  # @2023-08-25: add
                return
            if engine_kwargs.get('pool_pre_ping') is not True:
                if not context.connection or context.sqlalchemy_exception.connection_invalidated:
                    now = datetime.now().timestamp()
                    last_connect_time = engine_err_info.get('connect_time')
                    if last_connect_time is None or now - last_connect_time > 1:  # interval >1s
                        engine_err_info['connect_time'] = now
                        engine.connect()  # reconnect

    except Exception:
        flaskz_logger.exception('Connect to database ' + database_uri + ' error\n')
        return

    if not is_app:
        return

    @app.teardown_appcontext
    def teardown_db(exception):
        """
        When the request ends, close the db session.
        :param exception:
        :return:
        """
        close_db_session()

    if app_config.get('FLASKZ_DATABASE_DEBUG') is True:
        @event.listens_for(Engine, "before_cursor_execute")
        def before_cursor_execute(conn, cursor, statement, parameters, context, executemany):
            conn.info.setdefault('_query_start_time', []).append(time.perf_counter())

        @event.listens_for(Engine, "after_cursor_execute")
        def after_cursor_execute(conn, cursor, statement, parameters, context, executemany):
            end_time = time.perf_counter()
            start_time = conn.info['_query_start_time'].pop(-1)
            append_debug_queries(Attribute(**{
                'start_time': start_time,
                'end_time': end_time,
                'duration': (end_time - start_time) * 1000,  # ms
                'statement': statement,
                'parameters': parameters,
                'context': context,
            }))

        db_slow_time = app_config.get('FLASKZ_DATABASE_DEBUG_SLOW_TIME', -1)  # set config['FLASKZ_DATABASE_DEBUG_SLOW_TIME'] to enable
        db_access_times = app_config.get('FLASKZ_DATABASE_DEBUG_ACCESS_TIMES', -1)  # set config['FLASKZ_DATABASE_DEBUG_ACCESS_TIMES'] to enable
        if db_slow_time > 0 or db_access_times > 0:
            @app.after_request
            def after_request(response):
                queries = get_debug_queries()
                times = len(queries)
                if times > 0:
                    if 0 < db_access_times < times:
                        flaskz_logger.warning('Query Times:\n' + str(times))

                    if db_slow_time > 0:
                        slow_queries = []
                        for query in queries:
                            if query.duration > db_slow_time:
                                slow_queries.append('--Duration: %fms\n--Statement:%s\n--Parameters: %s' % (query.duration, query.statement, query.parameters))
                        if len(slow_queries) > 0:
                            flaskz_logger.warning('Slow queries:\n' + ('\n'.join(slow_queries)))
                return response


from ._base import *
from ._model import *
from ._util import *
from ._query_util import *

from ..utils._cls import Attribute
from ..utils._app import get_app_config_items
from ..log import flaskz_logger



from sqlalchemy import desc, asc, Column
from sqlalchemy.sql.elements import UnaryExpression

__all__ = ['parse_pss']


def parse_pss(cls, pss_config=None):
    """
    Get the search, sort and paging information.
    flaskz.utils.get_pss == flaskz.models.parse_pss
    pss = page+search+sort

    .. versionadded:: 1.6.1 - rename flaskz.utils.get_pss function to flaskz.models.parse_pss, flaskz.utils.get_pss is still available
    .. versionupdated:: 1.6.1 - change return list item(avoid SQL injection): SQL text --> Column.operator(parameter)  ex) "name like '%admin%'" --> TemplateModel.name.like('%admin%')

    Example:
        result, result = TemplateModel.query_pss(get_pss(   # use flaskz.utils.get_pss to format condition
            TemplateModel, {   # FROM templates
                "search": {                         # WHERE
                    "like": "t",                    # name like '%t%' OR description like '%t%' (TemplateModel.like_columns = ['name', description])
                    "age": {                        # AND (age>1 AND age<20)
                        ">": 1,                     # operator:value, operators)'='/'>'/'<'/'>='/'<='/'BETWEEN'/'LIKE'/'IN'
                        "<": 20
                    },
                   "city": {                        # AND (city IN ('Paris','London'))
                        "in": ['Paris' ,'London'],
                    },
                    "email": "taozh@focus-ui.com",  # AND (email='taozh@focus-ui.com')
                    "_ors": {                       # AND (country='America' OR country='Canada')
                        "country": "America||Canada"
                    },
                    "_ands": {                      # AND (grade>1 AND grade<5)
                        "grade": {
                            ">": 1,
                            "<": 5
                        }
                    }
                },
                "sort": {                           # ORDER BY templates.name ASC
                    "field": "name",
                    "order": "asc"
                },
                # "sort":[                            # multiple sort
                #     {"field": "name", "order": "asc"},
                #     {field": "age", "order": "desc"}
                # ],
                "page": {                           # LIMIT ? OFFSET ? (20, 0)
                    "offset": 0,
                    "size": 20
                }
            }))

    :param cls: Model class
    :param pss_config: page+search+sort
    :return:
    """
    pss_config = get_dict(pss_config)
    # --------------------search--------------------
    search = pss_config.get('search') or {}

    likes = _get_search_like_filters(cls, search.pop('like', None), search.pop('ilike', None))

    ands = []
    ors = []
    _ands = search.pop('_ands', None)
    if _ands:
        for key in _ands:
            _append_search_filters(ands, cls, key, _ands[key])

    _ors = search.pop('_ors', None)
    if _ors:
        for key in _ors:
            _append_search_filters(ors, cls, key, _ors[key])

    for key in search:
        _append_search_filters(ands, cls, key, search[key])

    # distinct = []  # @2023-06-07 add
    # _distinct = search.pop('_distinct', None)
    # if _distinct:
    #     if is_str(_distinct):
    #         _distinct = [_distinct]
    #
    #     if is_list(_distinct):
    #         for distinct_item in _distinct:
    #             distinct_column = cls.get_column_by_field(distinct_item)
    #             if distinct_column is not None:
    #                 distinct.append(distinct_column)

    # --------------------group--------------------
    groups = _get_groups(cls, pss_config.get('group', None))  # @2023-06-07 add

    # --------------------page--------------------
    page = pss_config.get('page') or {}
    offset = page.get('offset') or page.get('skip') or 0
    limit = page.get('limit') or page.get('size') or 100000

    # --------------------sort--------------------
    # @2022-04-10 fix exception subs2 = relationship('PerfTestSubModel2', cascade='all,delete-orphan', lazy='joined') ->
    # Can't resolve label reference for ORDER BY / GROUP BY / DISTINCT etc. Textual SQL expression 'f2' should be explicitly declared as text('f2')
    orders = _get_sorts(cls, pss_config.get('sort', None))

    return {
        'filter_ands': _filter_pss_list(ands),
        'filter_ors': _filter_pss_list(ors),
        'filter_likes': _filter_pss_list(likes),

        'order': _filter_pss_list(orders),
        'group': _filter_pss_list(groups),

        'offset': offset,
        'limit': limit
        # 'distinct': distinct
    }


def _filter_pss_list(items):
    return [item for item in items if item is not None]


def _get_groups(cls, group):
    """
    Get group by list
    """
    groups = []
    if group:
        if is_str(group):
            group = [group]
        if is_list(group):
            for group_item in group:
                group_column = cls.get_column_by_field(group_item)
                if group_column is not None:
                    groups.append(group_column)
    return groups


def _get_sorts(cls, sort):
    """
    Get sort list.
    """
    orders = []
    if is_list(sort):
        sorts = sort
    else:  # @2023-08-31 elif is_dict(sort):-->else
        sorts = [sort]

    for sort_item in sorts:
        if sort_item is None or sort_item == '':
            continue

        if is_str(sort_item):  # sort: "name"/["name"...]
            sort_column = cls.get_column_by_field(sort_item)
            if sort_column is not None:
                orders.append(asc(sort_column))
        elif type(sort_item) is Column:  # User.name
            if sort_item in cls.get_columns():
                orders.append(asc(sort_item))
        elif type(sort_item) is UnaryExpression:  # desc(User.name)/asc(User.name)
            orders.append(sort_item)
        elif is_dict(sort_item):  # {"field": "name", "order": "desc"}
            _order = sort_item.get('order')
            if _order and not is_str(_order):  # {"order": desc(User.name))} == "sort":desc(User.name)
                orders.append(_order)
            else:
                sort_field = sort_item.get('field')  # {"field": "name", "order": "asc"}
                if sort_field:
                    sort_column = cls.get_column_by_field(sort_field)
                    if sort_column is not None:
                        if _order and str(_order).strip().lower() in ['desc', 'descend', 'descending']:  # {"field": "name", "order": "desc"}
                            orders.append(desc(sort_column))
                        else:
                            orders.append(asc(sort_column))  # {"field": "name"}.
    return orders


def _get_search_like_filters(cls, search_like, search_ilike):
    """
    Get like list by like_columns.
    """
    like_filters = []

    like_columns = []
    for col in getattr(cls, 'like_columns', []):
        if is_str(col):
            col = cls.get_column_by_field(col)
        if col is not None:
            like_columns.append(col)
    if len(like_columns) == 0:
        return like_filters

    if search_like is not None and search_like != '':
        if not (search_like.startswith('%') or search_like.startswith('%')):
            search_like = "%" + search_like + "%"
        for col in like_columns:
            like_filters.append(get_col_op(col, 'like', search_like))

    if search_ilike is not None and search_ilike != '':
        if not (search_ilike.startswith('%') or search_ilike.startswith('%')):
            search_ilike = "%" + search_ilike + "%"
        for col in like_columns:
            like_filters.append(get_col_op(col, 'ilike', search_ilike))
    return like_filters


def _append_search_filters(items, cls, key, value):
    """
    Append filter item
    """
    col = cls.get_column_by_field(key)
    if col is None:
        return items

    if value is not None:
        if is_str(value):
            value = value.strip()
            if value != '':
                val_arr = value.split('||')
                for op_v in val_arr:
                    items.append(get_col_op(col, '==', op_v))
        elif is_dict(value):
            for operator, op_v in value.items():
                items.append(get_col_op(col, operator, op_v))
        else:
            items.append(get_col_op(col, '==', value))
    return items


def get_col_op(column, operator, value):
    """
        < , <= , == , != , > , >=
        in , notin , between ,
        is , isnot
        like , ilike , notlike , notilike

        ~startswith , endswith , contains~
    """
    operator = operator.lower()
    op_func_mapping = {
        'in': lambda col, val: col.in_(val) if type(val) is list else None,
        'notin': lambda col, val: col.notin_(val) if type(val) is list else None,
        'between': lambda col, val: col.between(*val) if type(val) is list else None,

        'is': lambda col, val: col.is_(val),
        'isnot': lambda col, val: col.is_not(val),

        'like': lambda col, val: col.like(val),
        'ilike': lambda col, val: col.ilike(val),
        'notlike': lambda col, val: col.notlike(val),
        'notilike': lambda col, val: col.notilike(val),

        '==': lambda col, val: col.__eq__(val)
    }
    if operator in op_func_mapping:
        return op_func_mapping.get(operator)(column, value)
    return column.op(operator)(value)


from ..utils import get_dict, is_str, is_dict, is_list



__all__ = ['get_list_args']


def get_list_args(params, args):  # @2022-05-11: add
    """
    Returns a single new list combining keys and args.

    def xxx(names, *args):
        print(get_list_args(names, args))

    xxx('a', ['b', 'c'])    -->['a', 'b', 'c']
    xxx('a', 'b', 'c')      -->['a', 'b', 'c']
    xxx(['a', 'b'], 'c')    -->['a', 'b', 'c']
    xxx(['a', 'b', 'c'])    -->['a', 'b', 'c']

    :param params:
    :param args:
    :return:
    """
    try:
        iter(params)
        if isinstance(params, (bytes, str)):
            params = [params]
        else:
            params = list(params)
    except TypeError:
        params = [params]
    if args and len(args) > 0:
        if isinstance(args[0], (tuple, list)):
            params.extend(*args)
        else:
            params.extend(args)
    return params



from sqlalchemy.exc import IntegrityError

from ._base import BaseModelMixin
from ._util import is_model_mixin_instance
from .. import res_status_codes
from ..log import flaskz_logger
from ..utils import is_dict

__all__ = ['ModelMixin']


class ModelMixin(BaseModelMixin):
    # -------------------------------------------add-------------------------------------------
    @classmethod
    def get_add_data(cls, data):
        """
        Return the added json data. By default, return the json data of the client
        Rewrite to customize the added data

        :param data: The data to be added
        :return:
        """
        return data

    @classmethod
    def before_add(cls, data):
        """
        The callback before adding data.
        If the return value is not True, the adding process will be terminated and the result will be returned to the client.

        :param data: The data to be added
        :return: True|Error Message
        """
        return True

    @classmethod
    def after_add(cls, data, instance, before_result):
        """
        The callback after adding data.
        If an exception occurs in after_add, the added data will not be rolled back.

        :param data: The data to be added
        :param instance: The result model instance, None if add fails.
        :param before_result: The result of the before_add callback
        :return:
        """
        pass

    @classmethod
    def add(cls, data):
        """
         Add the data to the db.
         Returns a tuple.
           --the first value represents whether the add operation was successful, True/False.
           --the second value represent the result of the add operation, reason/instance.

        Example:
            success, ins = User.add({"name": "taozh", "email": "taozh@focus-ui.com"})

        :param data:
        :return:
        """
        instance = None
        try:
            data = cls.get_add_data(data)
            check_result = cls.check_add_data(data)
            if check_result is not True:  # ex)db_data_already_exist
                return False, check_result

            before_result = cls.before_add(data)
            try:
                if before_result is True:
                    instance = cls.add_db(data)
            finally:
                cls.after_add(data, instance, before_result)  # after_add execute regardless of whether the addition is successful
        except Exception as e:
            flaskz_logger.exception(e)
            return False, res_status_codes.db_add_err

        return _op_result(before_result, instance)

    # -------------------------------------------update-------------------------------------------
    @classmethod
    def get_update_data(cls, data):
        """
        Return the updated json data. By default, return the json data of the client
        Rewrite to customize the updated data

        :param data: The updated json data
        :return:
        """
        return data

    @classmethod
    def before_update(cls, data):
        """
        The callback before updating data.
        If the return value is not True, the update process will be terminated and the result will be returned to the client.

        :param data:
        :return:
        """
        return True

    @classmethod
    def after_update(cls, data, instance, before_result):
        """
        The callback after updating data.
        If an exception occurs in after_update, the updated data will not be rolled back.

        :param data:
        :param instance:
        :param before_result:
        :return:
        """
        pass

    @classmethod
    def update(cls, data):
        """
        Update the data to the db.
        Returns a tuple.
           --the first value represents whether the update operation was successful, True/False.
           --the second value represent the result of the update operation, reason/instance.

        Example:
            success, ins = User.update({"id": 1, "email": "taozh@focus-ui.com"})

        :param data:
        :return:
        """
        instance = None
        try:
            data = cls.get_update_data(data)
            check_result = cls.check_update_data(data)
            if check_result is not True:  # ex)db_data_not_found / db_data_already_exist
                return False, check_result

            before_result = cls.before_update(data)
            try:
                if before_result is True:
                    instance = cls.update_db(data)
            finally:
                cls.after_update(data, instance, before_result)
        except Exception as e:
            flaskz_logger.exception(e)
            return False, res_status_codes.db_update_err

        return _op_result(before_result, instance)

    # -------------------------------------------delete-------------------------------------------
    @classmethod
    def get_delete_data(cls, pk_value):
        """
        Return the primary key of the data to be deleted. By default, return the primary key of the client data.
        Rewrite to customize the deleted data.

        :param pk_value: The primary key of the data to be deleted
        :return:
        """
        return pk_value

    @classmethod
    def before_delete(cls, pk_value):
        """
        The callback before deleting data.
        If the return value is not True, the delete process will be terminated and the result will be returned to the client.

        :param pk_value:
        :return:
        """
        return True

    @classmethod
    def after_delete(cls, pk_value, instance, before_result):
        """
        The callback after deleting data.
        If an exception occurs in after_delete, the data will still be deleted.

        :param pk_value:
        :param instance:
        :param before_result:
        :return:
        """
        pass

    @classmethod
    def delete(cls, pk_value):
        """
        Delete the data from the db.
        Returns a tuple.
           --the first value represents whether the delete operation was successful, True/False.
           --the second value represent the result of the delete operation, reason/instance.

        Example:
            success, ins =  User.delete(1)
            success, ins =  User.delete({'id': 1})

        :param pk_value:
        :return:
        """
        if is_dict(pk_value):  # {id:10}
            pk_value = cls._get_pk_value(pk_value)

        instance = None
        try:
            pk_value = cls.get_delete_data(pk_value)
            check_result = cls.check_delete_data(pk_value)
            if check_result is not True:  # ex)db_data_not_found
                return False, check_result

            before_result = cls.before_delete(pk_value)
            try:
                if before_result is True:
                    instance = cls.delete_db(pk_value)
            finally:
                cls.after_delete(pk_value, instance, before_result)
        except IntegrityError as e:
            flaskz_logger.exception(e)
            return False, res_status_codes.db_data_in_use
        except Exception as e:
            flaskz_logger.exception(e)
            return False, res_status_codes.db_delete_err

        return _op_result(before_result, instance)

    # -------------------------------------------query-------------------------------------------
    @classmethod
    def query_all(cls):
        """
        Override the base query_all method and return success flag.
        Used in router to return query data.

        Example:
            success, ins_list = User.query_all()

        :return:
        """
        try:
            return True, super().query_all()
        except Exception as e:
            flaskz_logger.exception(e)
            return False, res_status_codes.db_query_err

    @classmethod
    def query_pss(cls, pss_option):
        """
        Override the base query_pss method and return success flag.
        Used in router to return query data.

        Example:
            result, ins_list = TemplateModel.query_pss(get_pss(   # use flaskz.utils.get_pss to format condition
                TemplateModel, {   # FROM templates
                    "search": {                         # WHERE
                        "like": "t",                    # name like '%t%' OR description like '%t%' (TemplateModel.like_columns = ['name', description])
                        "age": {                        # AND (age>1 AND age<20)
                            ">": 1,                     # operator:value, operators)'='/'>'/'<'/'>='/'<='/'BETWEEN'/'LIKE'/'IN'
                            "<": 20
                        },
                        "email": "taozh@focus-ui.com",  # AND (email='taozh@focus-ui.com')
                        "_ors": {                       # AND (country='America' OR country='Canada')
                            "country": "America||Canada"
                        },
                        "_ands": {                      # AND (grade>1 AND grade<5)
                            "grade": {
                                ">": 1,
                                "<": 5
                            }
                        }
                    },
                    "sort": {                           # ORDER BY templates.name ASC
                        "field": "name",
                        "order": "asc"
                    },
                    # "sort":[                          # ORDER BY templates.name ASC, templates.age DESC
                    #     {"field": "name", "order": "asc"},
                    #     {field": "age", "order": "desc"}
                    # ],
                    "page": {                           # LIMIT ? OFFSET ? (20, 0)
                        "offset": 0,
                        "size": 20
                    }
                }))

        :param pss_option:
        :return:
        """
        try:
            return True, super().query_pss(pss_option)
        except Exception as e:
            flaskz_logger.exception(e)
            return False, res_status_codes.db_query_err


def _op_result(before_result, instance):
    if before_result is not True:
        return False, before_result
    if not is_model_mixin_instance(instance):
        return False, instance
    return True, instance



import math
import time

from flask import current_app, g

from ._common import is_dict

__all__ = ['set_app_cache', 'get_app_cache', 'clear_app_cache', 'set_g_cache', 'get_g_cache', 'remove_g_cache']


def _generate_cache_expire_data(data, expire_minutes=0):
    if type(expire_minutes) == int and expire_minutes > 0:
        data = {
            '_zexpires_time': math.floor(expire_minutes * 60 + time.time()),
            'data': data,
        }
    return data


def set_app_cache(key, data, expire_minutes=0):
    """
    Set the current application data cache, such as menu items.

    Example:
        set_app_cache('sys_module_name_mapping', module_name_mapping)

    :param expire_minutes:
    :param key:
    :param data:
    :return:
    """
    if current_app:
        cache = getattr(current_app, 'z_data_cache', None)
        if cache is None:
            cache = current_app.z_data_cache = {}
        cache[key] = _generate_cache_expire_data(data, expire_minutes)


def get_app_cache(key):
    """
    Get the current application data cache by the key.

    Example:
        module_name_mapping = get_app_cache('sys_module_name_mapping')

    :param key:
    :return:
    """
    if current_app:
        cache = getattr(current_app, 'z_data_cache', None)
        if cache:
            data = cache.get(key)
            if is_dict(data) and '_zexpires_time' in data:
                if data.get('_zexpires_time') < time.time():
                    return None
                return data.get('data')
            return data

    return None


def clear_app_cache():
    """
    Clear all the current application data caches.

    :return:
    """
    if current_app:
        cache = getattr(current_app, 'z_data_cache', None)
        if cache:
            cache.clear()


def set_g_cache(key, data):
    """
    Set the data cache in g, such as db session.

    Example:
        set_g_cache('_flaskz_db_session', session)

    :param key:
    :param data:
    :return:
    """
    if g:
        cache = getattr(g, 'z_data_cache', None)
        if cache is None:
            cache = g.z_data_cache = {}
        cache[key] = data


def get_g_cache(key):
    """
    Get the g data cache by the key.

    Example:
        session = get_g_cache('_flaskz_db_session')

    :param key:
    :return:
    """
    if g:
        cache = getattr(g, 'z_data_cache', None)
        if cache:
            return cache.get(key)
        return None


def remove_g_cache(key):
    """
    Remove the g data cache.

    :param key:
    :return:
    """
    if g:
        cache = getattr(g, 'z_data_cache', None)
        if cache:
            del cache[key]



import datetime
import json
import logging
import os
from logging.handlers import TimedRotatingFileHandler
from typing import Union

from werkzeug.local import LocalProxy

_flaskz_logger: logging.Logger = None  # @2022-04-27: fix _get_app_logger --> NameError: name '_flaskz_logger' is not defined
flaskz_logger: Union[LocalProxy, logging.Logger] = LocalProxy(lambda: _get_app_logger())  # @2023-05-15: add type Union


def init_log(app):
    """
    Initialize system log configuration

    Example:
        init_log(app)

    :param app:
    :return:
    """
    app_config = get_app_config_items(app) or {}
    level = logging.getLevelName(app_config.get('FLASKZ_LOGGER_LEVEL'))
    formatter = logging.Formatter(app_config.get('FLASKZ_LOGGER_FORMAT'))
    global _flaskz_logger
    _flaskz_logger = logging.getLogger('flaskz_logger')
    _flaskz_logger.setLevel(level)

    log_handler = app_config.get('FLASKZ_LOGGER_HANDLER')
    if log_handler is None:
        filename = app_config.get('FLASKZ_LOGGER_FILENAME')

        if filename is not None:
            filepath = os.path.join(app_config.get('FLASKZ_LOGGER_FILEPATH') or os.path.join(os.getcwd(), './syslog'), filename)
            log_handler = TimedRotatingFileHandler(
                encoding='utf-8',
                filename=filepath,
                when=app_config.get('FLASKZ_LOGGER_WHEN'),
                interval=1,
                backupCount=int(app_config.get('FLASKZ_LOGGER_BACKUP_COUNT')))
        else:
            log_handler = logging.StreamHandler()

    log_handler.setLevel(level)
    log_handler.setFormatter(formatter)
    _flaskz_logger.addHandler(log_handler)
    _flaskz_logger.disabled = app_config.get('FLASKZ_LOGGER_DISABLED') is True

    wz_logger = logging.getLogger('werkzeug')
    wz_logger.disabled = app_config.get('FLASKZ_WZ_LOGGER_DISABLED') is True


def clear_log_handler():
    logger = _get_app_logger()
    for handler in list(logger.handlers):
        logger.removeHandler(handler)


def _get_app_logger():
    return _flaskz_logger or logging.getLogger('flaskz_logger')


def get_log_data(data):
    return json.dumps(data, default=_log_data_converter)


def _log_data_converter(value):
    if isinstance(value, datetime.datetime):
        return value.__str__()


from ..utils._app import get_app_config_items
